#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : __init__.py.py
@Created       : 26/12/2017
"""

from aiospider.tools.singleton import OnlySingleton


class AIOSpider(metaclass=OnlySingleton):
    def __init__(self, loop):
        self.config = dict()
        self.loop = loop

    def __eq__(self, other):
        return id(self.config) == id(other.config)

    def __repr__(self):
        return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : test_pytest.py
@Created       : 26/12/2017
"""
import asyncio
import time

from aiospider import AIOSpider
from aiospider.tools.bloom_tools import BloomFilter
from aiospider.tools.job_queue import JobQueue
from aiospider.tools.redis_pool import RedisPool
from aiospider.workers import Worker

loop = asyncio.get_event_loop()


def test_config():
    app = AIOSpider(loop)
    app.config['redis_conn'] = 'redis://localhost:6379/2'
    assert app.config['redis_conn'] == 'redis://localhost:6379/2'


def test_redis():
    app = AIOSpider(loop)
    app.config['redis_conn'] = 'redis://localhost:6379/2'
    pool = RedisPool(app=app)
    pool3 = RedisPool(app=app)

    async def test_pool():
        pool2 = await pool.pool
        pool4 = await pool3.pool
        assert pool2 == pool4
        # res = await pool2.execute('bf.add', 'test', '1')

    # loop = asyncio.get_event_loop()
    loop.run_until_complete(test_pool())
    # loop.close()


def test_queue():
    app = AIOSpider(loop)
    app.config['redis_conn'] = 'redis://localhost:6379/2'
    # rp = RedisPool(app=app)
    jq = JobQueue(app, 'test')

    async def test_pool():
        print(await jq.push('1'))
        assert await jq.get() == b'1'

    # loop = asyncio.get_event_loop()
    loop.run_until_complete(test_pool())
    # loop.close()


def test_bloom():
    app = AIOSpider(loop)
    app.config['redis_conn'] = 'redis://localhost:6379/2'
    # rp = RedisPool(app=app)
    bf = BloomFilter(app, 'test_bloom')
    s = str(int(time.time()))

    async def test_pool():
        print(await bf.add(s))
        print(await bf.exists(s))
        assert await bf.exists(s) == 1

    loop.run_until_complete(test_pool())
    # loop.close()


def test_worker():
    app = AIOSpider(loop)
    queue = Worker(app).req_queue
    print(queue)


def test_aiospider():
    app = AIOSpider(loop)
    app2 = AIOSpider(2)
    assert app == app2


def test_job():
    app = AIOSpider(loop)
    app.config['redis_conn'] = 'redis://localhost:6379/2'
    # rp = RedisPool(app=app)
    bf = BloomFilter(app, 'test_bloom')
    s = str(int(time.time()))

    async def test_pool():
        print(await bf.add(s))
        print(await bf.exists(s))
        assert await bf.exists(s) == 1

    loop.run_until_complete(test_pool())


if __name__ == '__main__':
    # test_config()
    # test_redis()
    test_worker()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : singleton.py
@Created       : 21/12/2017
"""


class OnlySingleton(type):
    def __init__(cls, name, bases, attrs):
        super(OnlySingleton, cls).__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(OnlySingleton, cls).__call__(*args, **kwargs)
        return cls._instance


class Singleton(type):
    def __init__(cls, name, bases, attrs):
        super(Singleton, cls).__init__(name, bases, attrs)
        cls._instance = {}

    def __call__(cls, *args, **kwargs):
        key = cls.key_gen(*args, **kwargs)
        cls._instance.setdefault(key, None)
        if cls._instance[key] is None:
            cls._instance[key] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instance[key]

    @staticmethod
    def key_gen(*args, **kwargs):
        return f'{args}:{kwargs.items()}'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : redis_pool.py
@Created       : 21/12/2017
"""
import aioredis

from aiospider import AIOSpider
from aiospider.tools.singleton import OnlySingleton


class RedisPool(metaclass=OnlySingleton):
    def __init__(self, app):
        self._app = app
        self._pool = None

    @property
    async def pool(self):
        if not self._pool:
            self._pool = await aioredis.create_connection(self.redis_url)
        return self._pool

    @property
    def app(self):
        if not self._app:
            self._app = AIOSpider(None)
        return self._app

    @property
    def redis_url(self):
        return self.app.config['redis_conn']


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : bloom_tools.py
@Created       : 22/12/2017
"""
from aiospider import AIOSpider
from aiospider.tools.redis_pool import RedisPool
from aiospider.tools.singleton import Singleton


class BloomFilter(metaclass=Singleton):
    ADD_COMMAND = 'BF.ADD'
    EXISTS_COMMAND = 'BF.EXISTS'

    def __init__(self, app, name):
        self._app = app
        self.name = name
        self.pool = None

    @property
    def app(self):
        if not self._app:
            self._app = AIOSpider(None)
        return self._app

    async def add(self, key):
        if not self.pool:
            self.pool = await RedisPool(self.app).pool
        return await self.pool.execute(self.ADD_COMMAND, self.name, key)

    async def exists(self, key):
        if not self.pool:
            self.pool = await RedisPool(self.app).pool
        return await self.pool.execute(self.EXISTS_COMMAND, self.name, key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : base_worker.py
@Created       : 22/12/2017
"""
import asyncio
import logging

from aiospider import AIOSpider
from aiospider.models.job import RequestJob, ResponseJob
from aiospider.tools.bloom_tools import BloomFilter
from aiospider.tools.job_queue import JobQueue


class Worker:
    keys = {}
    name = 'base'
    worker = 'base'
    url_regex = None
    method = None
    params = {}
    require_params = []
    headers = {}

    def __init__(self, app):
        self._app = app

    @property
    def app(self):
        if not self._app:
            self._app = AIOSpider(None)
        return self._app

    async def analyze(self, req_job, content):
        # do something with request success
        pass

    @staticmethod
    def parser(data: dict):
        return data

    @staticmethod
    def pre_handle(resp_job: ResponseJob):
        return resp_job.success, resp_job.content

    async def fail(self, request_job):
        # do something with request fail
        pass

    async def _per_handle(self, resp_job: ResponseJob):
        success, content = self.pre_handle(resp_job)
        if success:
            return await self.analyze(resp_job.request_job, content=self.parser(content))
        else:
            await self.fail(resp_job)

    @classmethod
    def request_builder(cls, keys: dict, params=None, headers=None, data=None, identity=None,
                        redirect_times=None,
                        allow_redirect=None,
                        proxy=None,
                        cookies=None):
        url = cls.url_regex
        for key, value in keys.items():
            url = cls.url_regex.replace('{%s}' % key, str(value))
        for must in cls.require_params:
            if must not in params:
                raise ValueError()
        if isinstance(params, dict):
            cls.params.update(params)
        if headers:
            cls.headers.update(headers)
        return RequestJob(worker=cls.worker, url=url, method=cls.method, params=cls.params, data=data,
                          identity=identity, headers=cls.headers,
                          keys=keys,
                          cookies=cookies, proxy=proxy, name=cls.name, allow_redirect=allow_redirect,
                          redirect_times=redirect_times)

    @classmethod
    def req_queue_name(cls):
        return f'{cls.name}:{cls.worker}:request'

    @property
    def req_queue(self):
        return JobQueue(self.app, name=self.req_queue_name())

    @classmethod
    def resp_queue_name(cls):
        return f'{cls.name}:{cls.worker}:response'

    @property
    def resp_queue(self):
        return JobQueue(self.app, name=self.resp_queue_name())

    @classmethod
    def filter_name(cls):
        return f'{cls.name}:{cls.worker}:filter'

    @property
    def filter(self):
        return BloomFilter(self.app, name=self.filter_name())

    async def start(self):
        print(f'{self.name}:{self.worker}:Start')
        while True:
            try:
                queue = self.resp_queue
                content = await queue.get()
                if content:
                    resp_job = ResponseJob.from_json(content.decode())
                    await self._per_handle(resp_job)
                else:
                    await asyncio.sleep(1)

            except Exception as e:
                logging.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author        : EINDEX Li
@File          : queue.py
@Created       : 21/12/2017
"""
from aiospider import AIOSpider
from aiospider.tools.redis_pool import RedisPool
from aiospider.tools.singleton import Singleton


class JobQueue(metaclass=Singleton):
    def __init__(self, app, name):
        self._app = app
        self.name = name
        self.pool = None

    @property
    def app(self):
        if not self._app:
            self._app = AIOSpider(None)
        return self._app

    async def push(self, *items, reverse=False):
        if not self.pool:
            self.pool = await RedisPool(self.app).pool
        # calling methods of Redis class
        command = f'{"r" if reverse else "l"}push'
        return await self.pool.execute(command, self.name, *items)

    async def get(self, reverse=False):
        if not self.pool:
            self.pool = await RedisPool(self.app).pool
        command = f'{"l" if reverse else "r"}pop'
        return await self.pool.execute(command, self.name)
